#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/time.h>

void* bench_main(void* arg)
{
	char** argv = arg;

	const int num_loop    = atoi(argv[2]);
	const int num_bytes   = atoi(argv[3]);
	const int num_connect = atoi(argv[4]);

	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr("192.168.20.2");
	//addr.sin_addr.s_addr = inet_addr("127.0.0.1");
	addr.sin_port = htons(9090);

	int fds[num_connect];
	int ret;

	for(int i=0; i < num_loop; ++i) {
		for(int i=0; i < num_connect; ++i) {
			fds[i] = socket(PF_INET, SOCK_STREAM, 0);
			if(fds[i] < 0) { perror("socket"); exit(1); }

			ret = connect(fds[i], (struct sockaddr*)&addr, sizeof(addr));
			if(fds[i] < 0) { perror("connect"); exit(1); }
		}

		int fd = fds[0];

		char* buf = malloc(num_bytes);
		char* p;
		char const* pend = buf + num_bytes;

		p = buf;
		do {
			ssize_t sent = write(fd, p, pend - p);
			if(sent <= 0) {
				if(sent != 0 && (errno == EAGAIN || errno == EINTR)) {
					continue;
				}
				perror("write");
				return NULL;
			}
			p += sent;
		} while(p < pend);

		p = buf;
		do {
			ssize_t bytes = read(fd, p, pend - p);
			if(bytes <= 0) {
				if(bytes != 0 && (errno == EAGAIN || errno == EINTR)) {
					continue;
				}
				perror("read");
				return NULL;
			}
			p += bytes;
		} while(p < pend);

		for(int i=0; i < num_connect; ++i) {
			close(fds[i]);
		}
	}

	return NULL;
}

int main(int argc, char* argv[])
{
	if(argc != 5) {
		printf("%s <#thread> <#loop> <#bytes> <#connect>\n", argv[0]);
		return 1;
	}

	signal(SIGPIPE, SIG_IGN);

	int nthread = atoi(argv[1]);
	pthread_t th[nthread];

	struct timeval start;
	gettimeofday(&start, NULL);

	for(int i=0; i < nthread-1; ++i) {
		//printf(">");
		pthread_create(&th[i], NULL, &bench_main, argv);
	}
	//printf(">\n");
	bench_main(argv);

	//printf("<");
	for(int i=0; i < nthread-1; ++i) {
		void* ret;
		pthread_join(th[i], &ret);
		//printf("<");
	}
	//printf("\n");

	struct timeval finish;
	gettimeofday(&finish, NULL);

	double sec = (finish.tv_sec - start.tv_sec)
		+ (double)(finish.tv_usec - start.tv_usec) / 1000 / 1000;

	printf("%f\n", sec);

	return 0;
}

