#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/uio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

#define ECHO_BUFFER_SIZE (1024*1024)

static bool do_echo(int fd)
{
	char buffer[ECHO_BUFFER_SIZE];
	while(true) {
		ssize_t bytes = read(fd, buffer, ECHO_BUFFER_SIZE);
		if(bytes <= 0) {
			if(bytes == 0) {
				return true;
			}
			perror("read");
			if(errno == EAGAIN || errno == EINTR) {
				continue;
			}
			return false;
		}

		const char* p = buffer;
		const char const* pend = p + bytes;
		do {
			ssize_t sent = write(fd, p, pend - p);
			if(sent <= 0) {
				if(bytes == 0) {
					return true;
				}
				if(errno == EAGAIN || errno == EINTR) {
					continue;
				}
				perror("write");
				return false;
			}
			p += sent;
		} while(p < pend);
	}
}

static int child_main(int lfd)
{
	//printf("forked.\n");
	retry:
	{
		int fd = accept(lfd, NULL, NULL);
		if(fd < 0) {
			goto retry;
		}

		int success = do_echo(fd);

		close(fd);

		return success ? 0 : 1;

	}
}

static int fork_process(int lfd)
{
	pid_t pid = fork();
	if(pid < 0) {
		return -1;
	}

	if(pid == 0) {
		int ret = child_main(lfd);
		exit(ret);
	}

	return 0;
}

int main(int argc, char* argv[])
{
	if(argc != 3) {
		printf("%s <port> <#process>\n", argv[0]);
		return 1;
	}

	int port = atoi(argv[1]);
	int num_process = atoi(argv[2]);

	int lfd = socket(PF_INET, SOCK_STREAM, 0);
	
	struct sockaddr_in addr_listen;
	addr_listen.sin_family = AF_INET;
	addr_listen.sin_addr.s_addr = INADDR_ANY;
	addr_listen.sin_port = htons(port);

	int on = 1;
	setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

	bind(lfd, (struct sockaddr*)&addr_listen, sizeof(addr_listen));
	listen(lfd, 2048);

	for(int i=0; i < num_process; ++i) {
		fork_process(lfd);
	}

	int stat;
	while(true) {
		wait(&stat);
		fork_process(lfd);
	}

	return 0;
}

