#include "asout.h"
#include <string.h>
#include <unistd.h>
#include <sys/epoll.h>
#include <sys/resource.h>
#include <pthread.h>

#define ASOUT_BACKLOG_SIZE 256

typedef struct asout_ctx_t {
	char* buffer;
	size_t count;
} asout_ctx_t;

struct asout_t {
	int epfd;
	asout_ctx_t* ctxvec;
};

asout_t* asout_new(void) {
	asout_t* out = (asout_t*)malloc(sizeof(asout_t));
	out->epfd = epoll_create(ASOUT_BACKLOG_SIZE);
	out->ctxvec = calloc(sizeof(asout_ctx_t), sysconf(_SC_OPEN_MAX));
	return out;
}

void asout_free(asout_t* out) {
	close(out->epfd);
	free(out->ctxvec);
	free(out);
}

int asout_get_fd(asout_t* out) {
	return out->epfd;
}

void asout_send(asout_t* out, int fd, const char* buffer, size_t count) {
	asout_ctx_t* ctx = &out->ctxvec[fd];

	if(ctx->buffer != NULL) {
		char* tmp = realloc(ctx->buffer, ctx->count+count);
		memcpy(tmp+ctx->count, buffer, count);
		ctx->buffer = tmp;
		ctx->count += count;

	} else {
		ctx->buffer = malloc(count);
		ctx->count  = count;
		memcpy(ctx->buffer, buffer, count);

		struct epoll_event ev;
		ev.events = EPOLLOUT;
		ev.data.fd = fd;
		epoll_ctl(out->epfd, EPOLL_CTL_ADD, fd, &ev);
	}
}

bool asout_handler(int fd, void* arg) {
	asout_t* out = (asout_t*)arg;

	struct epoll_event log[ASOUT_BACKLOG_SIZE];
	int num = epoll_wait(out->epfd, log, ASOUT_BACKLOG_SIZE, -1);

	for(int i = 0; i < num; ++i) {
		int fd = log[i].data.fd;
		asout_ctx_t* ctx = &out->ctxvec[fd];

		ssize_t bytes = write(fd, ctx->buffer, ctx->count);
		if(bytes <= 0 || (ctx->count -= bytes) == 0) {
			free(ctx->buffer);
			ctx->buffer = NULL;
			epoll_ctl(out->epfd, EPOLL_CTL_DEL, fd, NULL);

		} else {
			struct epoll_event ev;
			ev.events = EPOLLOUT;
			ev.data.fd = fd;
			epoll_ctl(out->epfd, EPOLL_CTL_MOD, fd, NULL);
		}
	}

	return true;
}

