#include "miniev.h"
#include <sys/epoll.h>
#include <fcntl.h>
#include <sys/resource.h>

#define MINIEV_BACKLOG_SIZE 1024

struct miniev_t {
	int epfd;
	miniev_cb_t* cbvec;
};

miniev_t* miniev_new(void) {
	miniev_t* mev = (miniev_t*)malloc(sizeof(miniev_t));
	mev->epfd = epoll_create(MINIEV_BACKLOG_SIZE);
	mev->cbvec =
		malloc(sizeof(miniev_cb_t) * sysconf(_SC_OPEN_MAX));
	return mev;
}

void miniev_free(miniev_t* mev) {
	close(mev->epfd);
	free(mev);
}

bool miniev_push(miniev_t* mev, int fd, miniev_cb_t cb) {
	fcntl(fd, F_SETFL, O_NONBLOCK);
	mev->cbvec[fd] = cb;
	struct epoll_event ev;
	ev.events = EPOLLIN;
	ev.data.fd = fd;
	return epoll_ctl(mev->epfd, EPOLL_CTL_ADD, fd, &ev) >= 0;
}


int miniev_next(miniev_t* mev, int timeout_msec) {
	struct epoll_event backlog[MINIEV_BACKLOG_SIZE];

	int num = epoll_wait(mev->epfd, backlog,
			MINIEV_BACKLOG_SIZE, timeout_msec);
	if(num <= 0) {
		return num;
	}

	for(int i=0; i < num; ++i) {
		int fd = backlog[i].data.fd;
		miniev_cb_t* cb = &mev->cbvec[fd];
		bool success = (*cb->handler)(fd, cb->arg);
		if(success) {
			continue;
		}
		epoll_ctl(mev->epfd, EPOLL_CTL_DEL, fd, NULL);
		(*cb->destructor)(cb->arg);
		close(fd);
	}

	return 0;
}

